""" Storage class for sets of pandas object.  Similar to a panel, 
    but does inherently store 3d data.  Data can be converted to 3d through
    methods, but otherwise is just a container."""

from collections import OrderedDict, Iterable
from copy import deepcopy

import logging
from skspec.logger import logclass
from skspec.plotting.multiplots import slice_plot
import skspec.core.utilities as put

logger = logging.getLogger(__name__) 


class MultiError(Exception):
    """ """

def mem_address(obj):
    """ Return memory address string for a python object.  Object must have
    default python object __repr__ (ie it would look something like:
        <pyparty.tools.grids.CartesianGrid object at 0x3ba2fb0>
    The address is merely returned by string parsing. """
    try:
        out = obj.__repr__().split()[-1]
    except Exception as E:
        raise Exception("Failed to return memory address by string parsing.  "
                         "Recieved following message: %s" % E.message)
    else:
        return out.strip("'").strip('>')
    

@logclass(log_name=__name__, public_lvl='debug')
class Stack(object):
    """ Base class to store pandas objects, with special operations to
    return as 3d data (eg panel) and to apply functions itemwise.  Items are
    stored in an ordered dict."""
    
    itemlabel = 'Item'
    
    _magic=['__len__', 
            '__iter__', 
            '__reversed__',
            '__contains__', 
            ]
    
    def __init__(self, data, keys=None, name='', sort_items=False):
        self.name = name
                                
        # Dictionary input
        if isinstance(data, dict):
            logger.debug('Initializing "%s" from dictionary.' % self.full_name)
            if sort_items:
                logger.debug('Sorting keys')
                self._data=OrderedDict(sorted(data.keys(), key=lambda t: t[0]))

            else:
                self._data=OrderedDict(data)

        else:
            
            if not isinstance(data, Iterable):
                logger.info('%s constructed from non-iterable... converting '
                            'data to an iterable' % self.full_name)
                data=[data]
                
            if keys:
                if not isinstance(keys, Iterable):
                    logger.info('%s constructed from non-iterable... converting '
                                'keys to an iterable' % self.full_name)
                    keys = [keys]

                if len(keys) != len(data):
                    raise ValueError('Length mistmatch: keys and data (%s,%s)'\
                        % (len(keys), len(data)))
                
            # If keys not passed, generate them    
            else:
                # Zipped data ((key, df), (key, df))               
                try:
                    keys, data = zip(*data)
                except Exception:                
                    keys=self._gen_keys(len(data))
                    if len(keys) > 1:
                        logger.warn("Generating keys %s-%s" % (keys[0], keys[-1]))
                    else:
                        logger.warn("Generating key %s" % keys[0])
                        
            self._data=OrderedDict( [ (key, data[i]) for (i, key) 
                                      in enumerate(keys) ])

        

    @property
    def _address(self):
        """ Property to make easily accesible by multicanvas """
        return mem_address(super(Stack, self).__repr__())
        
      
    def _gen_keys(self, length):
        """ Return a list of itemlables (item0, item1 etc...) using
            self.itemlabel and a length"""

        logger.debug('Items not found on %s: generating item list' % self.full_name)
        return [self.itemlabel+str(i) for i in range(length)]                  
                
    # --------------------
    # Dictionary Interface  
    def __getitem__(self, keyslice):
        """ If single name, used dict interface.  If slice or integer, uses 
        list interface.  All results parameterized to key, data pairs, passed
        directly into a new Stack.
        """
        # Slice as list of strings or int [0, 'foo', 2, 'bar']
        if hasattr(keyslice, '__iter__'):

            tuples_out = []
            for item in keyslice:
                if isinstance(item, str):
                    item = self._data.keys().index(item)
                tuples_out.append(self._data.items()[item])
                        
        else:
            if isinstance(keyslice, int) or isinstance(keyslice, slice):
                tuples_out = self._data.items()[keyslice] 
            else: 
                tuples_out = [(keyslice, self._data[keyslice])]  #keyslice is name              
        
        # If single item, return TimeSpectra, else, return new Stack
        # Canonical slicing implementaiton; don't change unless good reason
        # Because len() wonky with nested tuples (eg (x,y) and [(x1,y1),(x2,y2)]
        # are both length two, this will work:
        
        if sum(1 for x in tuples_out) == 2:
            return tuples_out[1] #Return timespectra
        else:
            return self.__class__(tuples_out)                


    def __delitem__(self, keyslice):
        """ Delete a single name, or a keyslice from names/canvas """        

        if isinstance(keyslice, str):
            idx = self.names.index(keyslice)        
            self.pop(idx)
        else:
            raise NotImplementedError("Deletion only supports single entry")

    def __setitem__(self, name, canvas):
        """ """
        if name in self.names:
            idx = self.names.index(name) 
            self.pop(idx)
            self.insert(idx, name, canvas)
            
        else:
            self.names.append(name)  
                
    def __getattr__(self, attr):
        """ If attribute not found, try attribute lookup in dictionary.  If
        that is not found, try finding attribute on self._data.
        
        For example, self.keys() will first look for self['keys'].  Since
        this isn't found, it calls self._data.keys().  But if I do 
        self.Item1, then it returns self['Item1'].  The very rare conflict
        case that a user has named the items a method that may already exist
        in the dictionary (eg items=['a','b','keys'] is addressed.
        """
        
        if attr in self._data.keys():
            if hasattr(self._data, attr):
                raise AttributeError('"%s attribute" found in both the items\
                and as a method of the underlying dictionary object.'%(attr))
            else:
                return self[attr]
        return getattr(self._data, attr)
        

    # Attributes deferred to self.data /dictionary
    def __len__(self):
        return self._data.__len__()
    
    def __iter__(self):
        return self._data.__iter__()
    
    def __reversed__(self):
        return self._data.__reversed__()
    
    def __contains__(self):
        return self._data.__contains__()

        
        
    def as_3d(self):
        """ Return 3d structure of data.  Default is panel."""
        raise Panel(data=self._data)
                 
        ### Data types without labels    

    
    #Is this realy necessary?  See pyparty.ParticleManger for possibly more consistent implementation
    def get_all(self, attr, astype=tuple):
        """Generator/tuple etc.. of (item, attribute) pairs. """
        
        return put._parse_generator(
            ((item[0], getattr(item[1], attr)) for item in self.items()), astype)
    
                
    def _get_unique(self, attr):
        """ Inspects Stack itemwise for an attribute for unique values.
        If non-unique value for the attributes are found, returns
        "mixed". 
        """
        unique = set(self.get_all(attr, astype=dict).values())
        if len(unique) > 1:
            return 'mixed'
        else:
            return tuple(unique)[0] #set doesn't support indexing         
        
        
    def set_all(self, attr, val, inplace=False):
        """ Set attributes itemwise.  
            If not inplace, returns new instance of self"""
        if inplace:
            for (key, item) in self.items():
                try:           
                    setattr(item, attr, val)    
                except Exception as E:
                    raise Exception('Could not set %s in "%s".  Received the following \
                     exception:\n "%s"'%(attr, key, E))
        else:
            out=deepcopy(self._data) #DEEPCOPY            
            for item in out:
                setattr(out[item], attr, val)                
            return self.__class__(out)


    def apply(self, func, *args, **kwargs):
        """ Applies a user-passed function, or calls an instance method itemwise.
        
        
        Parameters:
        -----------
        func: str or function
            If string, must correspond to a method on the object stored 
            itemwise in the stack.  If a function, appliked itemwise to
            objects stored.  
              
        inplace: False 
            Special kwarg.  If true, self._data modified inplace, 
            otherwise new specstack is returned.
                         
        *args, **kwargs: 
            func arguments.
          
        Returns:
        --------
        If not inplace, returns SpecStack after itemwise application.
            
        """   

        inplace=kwargs.pop('inplace', False)
        
        if isinstance(func, basestring):
            if inplace:
                for item in self:
                    self[item] = getattr(self[item], func)(*args, **kwargs)      

            else:                
                return self.__class__(OrderedDict([(k, getattr(v, func)(*args, \
                                 **kwargs)) for k,v in self.items()]))                
             
        # function, numpyfunction etc...   
        else:
            if inplace:
                for item in self:
                    self[item] = self[item].apply(func)(*args, **kwargs)
                    
            else:
                return self.__class__(OrderedDict([(k, v.apply(func, *args, \
                                 **kwargs)) for k,v in self.items()]))                  
                
        

    @property  
    def full_name(self):
        """ Timespectra:name or Timespectra:unnamed.  Useful for scripts mostly """
        outname = getattr(self, 'name', 'unnamed')
        return '%s:%s' % (self.__class__.__name__, self.name)           
    
    #def __repr__(self):
        #""" """



@logclass(log_name=__name__)               
class SpecStack(Stack):
    """ Stack for just storing timespectra objects."""        

    itemlabel='spec_'

    def as_3d(self, **kwargs):
        """ Returns a 3d stack (SpecPanel) of the currently stored items.
            Additional kwargs can be passed directly to SpecPanel constructor."""
        from specpanel import SpecPanel      
        return SpecPanel(data=self._data, **kwargs)        
    
    
    ### Special properties for swift, in-place attribute overwrites of most 
    ### common itemwise operation.  Getter only tests for uniqueness
    
    @property
    def specunit(self):
        return self._get_unique('specunit')

                
    ### Do I want to make as a _set method to avoid accidental overwrite?
    @specunit.setter
    def specunit(self, unit):
        """ Sets specunit for every stored TimeSpectra."""
        self.set_all('specunit', unit, inplace=True)
    
    @property
    def iunit(self):
        return self._get_unique('iunit')    
    
    @iunit.setter
    def iunit(self, unit):
        """ Sets iunit for every stored TimeSpectra."""
        self.set_all('iunit', unit, inplace=True)    
        
    @property
    def reference(self):
        return self._get_unique('reference')
    
    @reference.setter
    def reference(self, ref):
        """ Set reference itemwise.  No getter, use get_all() instead."""
        self.set_all('reference', ref, inplace=True)

    ### This shouldn't have a setter
    @property
    def varunit(self):
        return self._get_unique('varunit')   
    
    def plot(self, *plotargs, **plotkwargs):
        """Returns multiplot of current stack.  
        
        Notes
        -----
        Wraps skspec.plotting.multiplots.slice_plot()
        """
        
        plotkwargs.setdefault('title', self.name)        
        if 'cbar' in plotkwargs:
            raise NotImplementedError("Colorbar on stack plot not yet supported")
        return slice_plot(self.values(), *plotargs, names=self.keys(), **plotkwargs)



    #def __repr__(self):
        #outstring = "%s (%s) %s: " % \
            #(self.__class__.__name__, self._address, 'what goes here')     
        #Ln = len(self)
        
        #if Ln == 0:                
            #outstring += 'EMPTY'

        #elif Ln >= MAXOUT:
            #outstring +=  '%s canvii (%s ... %s)' % \
                #(Ln, self.names[0], self.names[-1])                        
            
        #else:
            #SEP_CHARACTER = '-'
            #_NEWPAD =  (PADDING-1) * ' '  # REduce CONFIG PADDING by one space
            #just_fcn = {
                #'l': str.ljust,
                #'r': str.rjust,
                #'c': str.center}[ALIGN]            
            
            #outstring += '\n'
            #outrows=[]
            #for idx, name in enumerate(self.names):
                #c = self.canvii[idx]
                #cx, cy = c.rez
                
                #col1 = '%s%s' % (_NEWPAD, name)
                #col2 = 'Canvas (%s) : %s X %s : %s particles' % \
                    #(c._address, cx, cy, len(c))
                #outrows.append([col1, SEP_CHARACTER, col2])
         
            #widths = [max(map(len, col)) for col in zip(*outrows)]
            #outstring = outstring + '\n'.join( [ _NEWPAD.join((just_fcn(val,width) 
                #for val, width in zip(row, widths))) for row in outrows] )

        #return outstring


        
   

if __name__=='__main__':
    ### For testing.
    from skspec.core.specindex import SpecIndex
    from skspec.core.timespectra import TimeSpectra
    from pandas import date_range, DataFrame, Index

    import numpy as np       
    import matplotlib.pyplot as plt
    from skspec.data import aunps_glass

    spec = SpecIndex(np.arange(400, 700,10), unit='nm' )
    spec2 = SpecIndex(np.arange(400, 700,10), unit='m' )
    
    testdates = date_range(start='3/3/12', periods=30, freq='h')
    testdates2 = date_range(start='3/3/12', periods=30, freq='h')
    
    #ts = TimeSpectra(abs(np.random.randn(30,30)), columns=testdates, index=spec)  
    #t2 = TimeSpectra(abs(np.random.randn(30,30)), columns=testdates2, index=spec2) 
    ts = aunps_glass(style=1)
    t2 = aunps_glass(style=2)
    d = (('d1', ts), ('d2', t2))

    ##x = Panel(d)
    stack = ts.as_iunit('r').split_by(5)
    stack.reference = 0
    #stack.plot(figsize=(8,8), colormap='RdBu')   
    
    #raise SyntaxError

    #def random_noise(curve):
        #from random import randint
        #if randint(1,5) == 5:     #20% chance of choosing 5
            #curve = curve + 25 * np.random.randn(len(curve))
        #return curve

    #out = y.apply(random_noise)
    #len(out)
    #print out
    #out.plot()
    #plt.show

    #y.reference = 0
    #y.iunit='a'
    #y.plot()
    #plt.show()
##    sys.exit()
    ##y[0:1]
    ##y['d1',1]
    ##x = y.set_all('specunit', 'cm')
    ##y._get_unique('specunit')
    ##y.apply('wavelength_slices', 8)
    ##y['d1']
    
    #x=y.apply(np.sqrt, axis=1)
    #print 'hi'
    ##y.applynew('boxcar', binwidth=10)
    ##y.applynew(boxcar, binwidth=10)
    ###print y.specunit
    ###y.specunit='f'
    ##print 'hi'


""" Wrappers and utilitiles for MPLD3 customization """

import mpld3
from mpld3.plugins import Reset, Zoom, BoxZoom, \
           PointLabelTooltip, PointHTMLTooltip, LineLabelTooltip, \
           MousePosition, LineHTMLTooltip

ALLPLUGINS = dict(
   reset=Reset,
   zoom=Zoom,
   boxzoom=BoxZoom, 
   pointlabel=PointLabelTooltip,
   pointHTMLlabel = PointHTMLTooltip, 
   linelabel=LineLabelTooltip, 
   linehtml=LineHTMLTooltip,
   mousepos=MousePosition
   )

# Reversed dictionary
ALLPLUGINS_REV = dict((v, k) for k,v in ALLPLUGINS.items())

# Plugins that require access to lines
LABELPLUGINS = [LineLabelTooltip, PointLabelTooltip]
HTMLPLUGINS = [LineHTMLTooltip, PointHTMLTooltip]

DEFAULTS = ('reset','boxzoom','zoom')

class PluginManagerError(Exception):
   """ """

class PluginManager(object):
   """ Wrapper for adding/removing/printing plugins available on a 
   mplfigure.  Returns a figure (via self.fig) with seleted plugins attached.
   Used primarily for widgets/guis where the state of plugins
   needs managed.  Otherwise, using MPLD3 API directly with a mpl figure
   is simple and preferred.
   """
   
   def __init__(self, fig, ax=None, names=None):
      """ """
 
      # Set fig logic through property
      self.reset_plugins()
      # Goes through property
      self.fig = fig
      
      self.ax = ax
      if self.ax is None:
         self.ax = fig.get_axes()[0]
         if len(fig.get_axes()) > 1:
            raise PluginManager('Multiple axes found in figure; pass axes directly!') 

      self.lines = self.ax.get_lines()
      if not self.lines:
         raise PluginManagerError('No lines found on axes; is this a line plot?') #should be warn

      # Set some names for lines
      self.names = names
      if not self.names:
         self.names = ['Line: %s' % i for i in range(len(self.lines))]
     
        
   def reset_plugins(self):
      """ Sets plugins to default values """
      self._plugins = dict((k,ALLPLUGINS[k]) for k in DEFAULTS
                           if k in ALLPLUGINS) 
   
   @property
   def fig(self):
      """ Returns figure with plugins attached."""
      mpld3.plugins.clear(self._fig)
      for p in self._plugins.values():
         if p in LABELPLUGINS:
            for idx, line in enumerate(self.lines):
               mpld3.plugins.connect(self._fig, p(*self.lines))            

         else:
            mpld3.plugins.connect(self._fig, p())
     #    except TypeError:
     #       mpld3.plugins.connect(self._fig, p(*self.lines))            
      return self._fig
                           
   @fig.setter
   def fig(self, fig):
      """ Upon new fig, sets default plugins from that figure. """
      if fig is None:
         raise PluginManager('FIG CANNOT BE NONE')
      
      self._fig = fig
      self._plugins = {}
      
      # By default, mpld3 adds plugins if finds mpl plot
      oldplugs = mpld3.plugins.get_plugins(fig)

      for plug in oldplugs:
         pclass = plug.__class__
         if pclass in ALLPLUGINS_REV:
            self.add(ALLPLUGINS_REV[pclass]) #Set key from class
         else:
            # fig has a plugin that we don't know about
            # (maybe should jsut add it with custom key like unknown plugin1
            raise PluginManagerError('UNKNOWN PLUGIN %s' % plug)
            
   @property
   def plugins(self):
      return self._plugins.values()
   
   @property
   def available(self):
      return self.ALLPLUGINS 
   
   def add(self, *plugins):
      """ Add a plugin to figure """
      for plugin in plugins:
         if not isinstance(plugin, basestring):
            raise PluginManagerError('Can only add plugins as strings; use dictionary'
                                  ' interface to add new plugins.')
         self._plugins[plugin] = ALLPLUGINS[plugin]
      
   def remove(self, *plugins):
      # Let dictionary do its own error handling...
      for plugin in plugins:
         del self._plugins[plugin]

   def __getitem__(self, *args, **kwargs):
      return self._plugins.__getitem__(*args, **kwargs)
   
   def __setitem__(self, *args, **kwargs):
      self._plugins.__setitem__(*args, **kwargs)
      
   def __delitem__(self, *args, **kwargs):
      self._plugins.__delitem__(*args, **kwargs)
      
#   def __repr__(self, *args, **kwargs):
#      """ """
 
if __name__ == '__main__':
   import matplotlib.pyplot as plt
   f, ax = plt.subplots()
   ax.plot([1,2,3,4])
   ax.plot([1,3,6,10])
   
#   points = ax.plot(range(10), 'o')
#   print points

#   points = ax.scatter(range(40), range(40))
   pman = PluginManager(fig=f)
   
   
   pman.add('mousepos')
   print pman._plugins
   pman.remove('boxzoom')
   pman.remove('zoom')
#   pman.add('linelabel')
   print pman._plugins
   
   

#   plugins.connect(fig, PointLabelTooltip(points[0]))
#   fig_to_html(fig)   
   mpld3.show(pman.fig)


import plot_utils as put
import skspec.core.utilities as pvutils
import skspec.config as pvcnfg
import logging
import numpy as np
import matplotlib.pyplot as plt

logger = logging.getLogger(__name__)

class PlotError(Exception):
    """ """

#XXX UPDATE DOCSTRING (HOW TO REFERENCE SPECPLOT TO THIS ONE)
def _genplot(ts, *args, **pltkwargs):
    """ Generic wrapper to ts._frame.plot(), that takes in x/y/title as parsed
    from various calling functions:
    NEW KEYWORDS:
        grid
        color
        labelsize
        titlesize
        ticksize 
        xlim/ylabel
        cbar
        ax
        fig
        xlabel
        ylabel
        title
    """
             
    # Add custom legend interface.  Keyword legstyle does custom ones, if pltkwrd legend==True
    # For now this could use improvement  
    xlabel = pltkwargs.pop('xlabel', '')
    ylabel = pltkwargs.pop('ylabel', '')
    title = pltkwargs.pop('title', '')

    pltkwargs.setdefault('legend', False)
    pltkwargs.setdefault('linewidth', 1)
    legstyle = pltkwargs.pop('legstyle', None)   
        
    # Adhere to cananoical "cmap" 
    if 'cmap' in pltkwargs:
        pltkwargs['colormap'] = pltkwargs.pop('cmap')    
    
    fig = pltkwargs.pop('fig', None)
    ax = pltkwargs.pop('ax', None)
    cbar = pltkwargs.pop('cbar', False)
    _barlabels = 5 #Number of ticks/labels in colorbar

    xlim = pltkwargs.pop('xlim', None)
    ylim = pltkwargs.pop('ylim', None)
    custompadding = pltkwargs.pop('custompadding', 0.05)
            
            
    if not ax:
        f, ax = plt.subplots(1)
        if not fig:
            fig = f
        
   
    # Grid (add support for minor grids later)
    grid = pltkwargs.pop('grid', 'black')
    
    labelsize = pltkwargs.pop('labelsize', pvcnfg.LABELSIZE) #Can also be ints
    titlesize = pltkwargs.pop('titlesize', pvcnfg.TITLESIZE)
    ticksize = pltkwargs.pop('ticksize', '') #Put in default and remove bool gate below

    pltkwargs['ax'] = ax            
    ax = ts._frame.plot(**pltkwargs)
    
    if cbar:
        if pltkwargs.get('color', None):
            raise PlotError('Colorbar requires cmap; solid color \
            "%s" found.' % pltkwargs['color'])

        c_rotation, c_reverse = 90, False
        if cbar in ['r', 'reverse']:
            c_rotation, c_reverse = 270, True
        if not fig:
            raise PlotError("Color bar requries access to Figure.  Either pass fig"
                            " keyword or do not pass custom AxesSubplot.")

        # Need a colormap if you have a colorbar!
        try:
            pcmap = pltkwargs['colormap']
        except KeyError:
            raise PlotError('Cannot plot a colorbar without a colormap!')

        mappable, vmin, vmax = put._annotate_mappable(ts, pcmap, axis=0)
        cbar = fig.colorbar(mappable, ticks=np.linspace(vmin, vmax, _barlabels))
        
        tunit = pvutils.safe_lookup(ts, 'full_varunit')
        
        cbar.set_label(r'%s$\rightarrow$' % tunit, rotation=c_rotation)
        
        if len(ts.columns) > _barlabels -1:
            label_indices = np.linspace(0, len(ts.columns), _barlabels)
            label_indices = [int(round(x)) for x in label_indices]
            if label_indices[-1] > len(ts.columns)-1:
                label_indices[-1] = len(ts.columns)-1 #Rounds over max
            
            labels = [ts.columns[x] for x in label_indices]

            # IF LABELS ARE FLOATS (NEED A BETTER CONDITION THAN THIS)
            try:
                labels = [round(float(x),put.float_display_units) for x in label_indices]
            except Exception:
                pass
        
        # Don't add custom labels if aren't at least 5 columns if DF        
        else:
            labels = []
            
        cbar.ax.set_yticklabels(labels)
            
        if c_reverse:
            cbar.ax.invert_yaxis()
        
    # Add minor ticks through tick parameters  
    ax.minorticks_on()
        
    ax.set_xlabel(xlabel, fontsize=labelsize)
    ax.set_ylabel(ylabel, fontsize=labelsize)
    ax.set_title(title, fontsize=titlesize)         
    
    # Not normazling padding correctly!
    
    def _correct_padding(xi,xf):
        """ Note, when making multiplots, this can be an issue and users
        will want to do padding=None
        """
        dlt_x = xf-xi
        boundary = abs(dlt_x *custompadding)
        low_bound = xi-boundary
        high_bound = xf+boundary
        return (low_bound, high_bound)
    
    
    if not xlim and custompadding is not None:
        try:
            xlim = _correct_padding(min(ts.index), max(ts.index))
            ax.set_xlim(xlim)
        # Padding not inferrable from string indicies like in time plots 
        except Exception:
            pass
                 
    if not ylim and custompadding is not None:
        try:
            ylim = _correct_padding(ts.min().min(), ts.max().max())
            ax.set_ylim(ylim)
        except Exception:
            pass
        
    
    if legstyle and pltkwargs['legend'] == True:  #Defaults to False
        if legstyle == 0:
            ax.legend(loc='upper center', ncol=8, shadow=True, fancybox=True)
        elif legstyle == 1:
            ax.legend(loc='upper left', ncol=2, shadow=True, fancybox=True)  
        elif legstyle == 2:
            ax=put.easy_legend(ax, position='top', fancy=True)

    # http://matplotlib.org/api/axis_api.html           
    # Other grid args like linestyle should be set directly with calls
    # to ax.grid(**linekwds)
    if grid:
        if grid == True:
            ax.grid()
        else:
            ax.grid(color=grid) #Let's any supported color in
        
    if ticksize:
        logger.info('Adjusting ticksize to "%s"' % ticksize)
        # Get all x and y ticks in a list
        allticks = ax.xaxis.get_majorticklabels()
        allticks.extend(  ax.yaxis.get_majorticklabels() )

        for label in allticks:
            label.set_fontsize(ticksize)
         #  label.set_fontname('courier')        

    return ax    


# Requires ranged dataframe used wavelength slices method
def range_timeplot(ranged_ts, **pltkwds):
    ''' Makes plots based on ranged time intervals from spec_utilities.wavelength_slices().
    Uses a special function, put._uvvis_colorss() to map the visible spectrum.  Changes default legend
    behavior to true.'''

    pltkwds['cmap'] = pltkwds.pop('cmap', 'jet')
    pltkwds['legend'] = pltkwds.pop('legend', True)
    pltkwds['linewidth'] = pltkwds.pop('linewidth', 2.0 )  
          
    tunit = pvutils.safe_lookup(ranged_ts, 'full_varunit')    
          
    pltkwds.setdefault('xlabel', tunit)     
    pltkwds.setdefault('ylabel', '$\int$ %s (sliced)' % ranged_ts.full_iunit)    
    pltkwds.setdefault('title', 'Area Ranges: '+ ranged_ts.name )       
                
    # Needs to be more robust and check specunit is index etc...
    return _genplot(ranged_ts.transpose(), **pltkwds)   #ts TRANSPOSE


def areaplot(ranged_ts, **pltkwds):
    """
    Makes plots based on ranged time intervals from spec_utilities.wavelength_slices().
    Uses a special function, put._uvvis_colorss() to map the visible spectrum.  
    Changes default legend behavior to true.
    
    Parameters:
    -----------
    ranged_ts: Spectra or Spectrum.  If n x m spectra, ts.area() will be called,
    and transposed().  If Spectrum, passed directlyt hrough.
    
    Notes:
    ------
    Added some extra functionality on 2/13/13, to make it ok if user passed in 
    transposed() or non-transposed(). Additionally, if user passes non Mx1
    dimensional plot, it recomputes the area.  Hence, user can just do areaplot(ts)
    """
    
    # If not M x 1 shape, recompute area
    if ranged_ts.ndim > 1:
        rows, cols = ranged_ts.shape
        if cols != 1 and rows != 1:
            try:
                ranged_ts = ranged_ts.area()
            except Exception:
                # Often error is if someone passed in transposed area using datetime index
                raise IOError('Could not successfully run .area() on %s object.' % 
                              type(ranged_ts))
            else:
                logger.warn('Recomputing area from shape (%s, %s) to %s' % 
                            (rows, cols, str(ranged_ts.shape)))

    # Replace w/ set defaults
    pltkwds.setdefault('legend', False)
    pltkwds.setdefault('linewidth', 3.0)
    pltkwds.setdefault('color', 'black') # If removing, colormap default in _genplot
                                         # Will cause bug

    tunit = pvutils.safe_lookup(ranged_ts, 'full_varunit')

    pltkwds.setdefault('xlabel', tunit)  
    pltkwds.setdefault('ylabel', '$\int$ %s d$\lambda$'%ranged_ts.full_iunit)    
    pltkwds.setdefault('title', 'Area: '+ ranged_ts.name )      


    # If shape is wrong, take transpose    
    if ranged_ts.ndim > 1:
        rows, cols = ranged_ts.shape    
        if cols == 1 and rows != 1:
            out = ranged_ts
        else:
            logger.warn("areaplot() forcing transpose on %s for plotting" 
                        % ranged_ts.name)
            # _df.transpose() causes error now because specunit is being transferred!
            out = ranged_ts.transpose()

    # If series, just pass in directly
    else:
        out = ranged_ts
                    
    # Pass in df.tranpose() as hack until I fix specunit (otherwise it tries to
    # assign specunit to the index, which is a time index (this causes error in new pandas)
    return _genplot(out, **pltkwds)   #ts TRANSPOSE
    
if __name__ == '__main__':
    from skspec.data import aunps_glass
    ts = aunps_glass()
    print ts.full_iunit
    ts.plot(kind='area')
#    areaplot(ts)
    plt.show()



import sys
import os.path as op
from setuptools import setup, find_packages

NAME = 'scikit-spectra'

# Python >= 2.7 ?
user_py = sys.version_info
if user_py < (2, 7):
    raise SystemExit('%s requires python 2.7 or higher (%s found).' % \
                     (NAME, '.'.join(str(x) for x in user_py[0:3])))


#XXX Get current module path and do path.join(requirements)
#with open(op.join(sys.path[0], 'requirements.txt')) as f:
with open('requirements.txt', 'r') as f:
    required = f.readlines()
    required = [line.strip() for line in required]
    
# For now, most of these are in here for testing.  Dependency version 
#requirements can probably be relaxed, especially chaco.
setup(
    name = NAME,
    version = '0.3.2-1',
    author = 'Adam Hughes',
    maintainer = 'Adam Hughes',
    maintainer_email = 'hughesadam87@gmail.com',
    author_email = 'hughesadam87@gmail.com',
    packages = find_packages(),

    # include .tex and .ipynb files
    package_data={
      'skspec.scripts.gwu_script.tex_templates': ['*.tex'],
      'skspec.scripts.gwu_script.mlab_templates': ['*.m'],   
#      'pyparty.bundled':['*.css'],   
      'skspec.examples.Notebooks':['*.ipynb'],
      'skspec.scripts.gwu_script':['*.png', '*.ipynb'],
      'skspec.data':['*'],
      'skspec.bundled':['*.css']
   },
       
    entry_points = {'console_scripts': 
                    [
                       'gwuspec = skspec.scripts.gwu_script.gwuspec:main',
                       'gwureport = skspec.scripts.gwu_script.gwureport:main'
                    ]
                    },
    
    url = 'http://hugadams.github.io/scikit-spectra/',
    download_url = 'https://github.com/hugadams/scikit-spectra',
    license = 'LICENSE.txt',
    description = 'Spectroscopy in Python built on Pandas.',
      classifiers = [
          'Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Developers',
          'Natural Language :: English',
#          'Operating System :: OSIndependent',  #This is limited only by 
          'Programming Language :: Python :: 2',
          'Topic :: Scientific/Engineering :: Visualization',
          'Topic :: Scientific/Engineering :: Chemistry',
          'Topic :: Scientific/Engineering :: Physics',
          'Topic :: Scientific/Engineering :: Medical Science Apps.',
          ],
 )



# CREDIT TO JAKE VANDERPLAUS MPLD3 FOR THIS EXAMPLE

import mpld3
from mpld3 import plugins, utils

# for color.  "stroke", "red" instead of "stroke-opacity", alpha-fg

#http://www.d3noob.org/2014/02/styles-in-d3js.html
class HighlightLines(plugins.PluginBase):
    """A plugin to highlight lines on hover"""

    JAVASCRIPT = """
    mpld3.register_plugin("linehighlight", LineHighlightPlugin);
    LineHighlightPlugin.prototype = Object.create(mpld3.Plugin.prototype);
    LineHighlightPlugin.prototype.constructor = LineHighlightPlugin;
    LineHighlightPlugin.prototype.requiredProps = ["line_ids"];
    LineHighlightPlugin.prototype.defaultProps = {alpha_bg:1.0, alpha_fg:0.1}
    function LineHighlightPlugin(fig, props){
        mpld3.Plugin.call(this, fig, props);
    };

    LineHighlightPlugin.prototype.draw = function(){
      for(var i=0; i<this.props.line_ids.length; i++){
         var obj = mpld3.get_element(this.props.line_ids[i], this.fig),
             alpha_fg = this.props.alpha_fg;
             alpha_bg = this.props.alpha_bg;
         obj.elements()
             .on("mouseover", function(d, i){
                            d3.select(this).transition().duration(50)
                              .style("stroke-opacity", alpha_fg); })
             .on("mouseout", function(d, i){
                            d3.select(this).transition().duration(200)
                              .style("stroke-opacity", alpha_bg); });
      }
    };
    """

    def __init__(self, lines):
        self.lines = lines
        self.dict_ = {"type": "linehighlight",
                      "line_ids": [utils.get_id(line) for line in lines],
                      "alpha_bg": 1.0, #How light to make highlighted line (can also set to alpha of original mpl plot this way)
                      "alpha_fg": 0.15} #How dark to make non-highlighted line



### Enthought tool suite imports 
from traits.api import Instance, Str, Enum, Range, HasTraits, Button, Enum, Property, Bool, Any
from traitsui.api import Item, View, HGroup, VGroup, Group, Include, ShellEditor
from enable.api import ComponentEditor
from chaco.api import LabelAxis, Plot, ToolbarPlot
from chaco.tools.api import BetterSelectingZoom, PanTool
from random import randint
from copy import deepcopy
from pandas import DataFrame

from chaco.tools.api import RangeSelection, RangeSelectionOverlay

### For testing 
import numpy as np
from time import sleep

### For skspec imports
#from skspec import get_csvdataframe
#from skspec import wavelength_slices
#from skspec import datetime_convert
from pandas.stats.moments import rolling_mean

### Local import 
from pandasplotdatav2 import PandasPlotData
   
    

### Add a method to relable rows/columns in place?
class PandasPlot(HasTraits):
    '''Plot to interface a pandas dataframe'''
    ### Actually   
    originaldata=Instance(DataFrame)
    dataframe=Instance(DataFrame)  
    primaryaxis=Enum(0,1,'index','columns') #Integer or string, defaults to 0
    _primaryaxis=Property(Enum(0,1), depends_on='primaryaxis') #0,1
    indexname=Str('Index')
    columnname=Str('Column')

    ### Traits of the actual plot and plotdatasource handler ###
    plot=Instance(Plot)  
    plotdata=Instance(PandasPlotData) #Data stored at any given time in the plot

    ### Aesthetic traits
    plot_title=Str('') #Defaults from dataframe name
    linecolor=Str('black')
    pointcolor=Str('red')
    linestyle=Enum('line', 'scatter', 'both')
    markersize=Range(low=1, high=5)

    ### Traits for testing    
    df_new=Button
    df_change=Button    
    re_bin=Range(1,100)

    def _dataframe_changed(self, old, new):
        ''' Handles how updates occur when dataframe changes.  Evaluates if columns or columnlabels
        have been changed.  Provides entry condition as well.

        Note: New automatically sets self.dataframe, so when I refer to new, I am actually
        referring to self.dataframe.  Using "new" instead of self.dataframe is just for readability'''

        ### Initialize plot first time dataframe is passed into the class.  Boolean listeners
        ### for dataframe behave oddly, so uses self.plotdata for entry condition.
        if not self.plotdata:
            self.plotdata=PandasPlotData(df=new)
            
            self.originaldata=new

            ### Try to infer plot title from dataframe name ###	    
            try: 
                self.plot_title=new.name
            except AttributeError:
                pass

            ### Draw barebones of plot
            self.draw_plot()
            ### Draw lines
            self._draw_lines() 

            return
        
        
        ### Decide to update columns or completely redraw dataframe.  
        else:
            labelold=self._getlabelarray(old)	    
            labelnew=self._getlabelarray(new)

            ### Have columns been added or removed?	    
            if len(labelold) != len(labelnew):
                self._overwrite_plotdata

            ### Has index along primaryaxis changed any?
            ### Pandas index comparison is a bit tricky so just list conver
            elif list(labelold) != list(labelnew):
                self._overwrite_plotdata
            else:
                print 'updating frame'
                self.plotdata.set_df(new)

    ### Properties ###	
    def _get__primaryaxis(self):
        ''' Converts 'columns' or 'index' to 0,1.  Mask basically'''
        if self.primaryaxis=='columns' or self.primaryaxis==0:
            return 0
        else:
            return 1

    def _getlabelarray(self, dataframe):
        ''' Conveience method to get label along working dimension of a dataframe.'''
        if self._primaryaxis==0:
            return np.asarray(dataframe.columns)
        else:
            return np.asarray(dataframe.index)

    def _df_change_fired(self):
        ''' set data iteratively column by column'''
        self.dataframe=DataFrame((np.random.randn(100,100)))

    def _df_new_fired(self):
        df2=DataFrame((np.random.randn(100,100)), columns=[i*5 for i in range(10)], index=[i*50 for i in range(10)])
        self.dataframe=df2
        self._draw_lines()
        
    def _re_bin_changed(self):
        ## Fix, squashing indicies
        #self.dataframe=rebin(self.dataframe, self.re_bin, axis=0, avg_fcn='mean')
        self.dataframe=rolling_mean(self.originaldata, self.re_bin)

    def _draw_lines(self):
        ''' Recasts plotdata and ties it to the plot object.'''
        for idx, name in enumerate(self.plotdata.list_data(as_strings=True)):
#            if idx % 50==0:
            self.plot.plot((self.indexname, name))        

    def draw_plot(self, toolbar=True, **pltkws):
        ''' Draw bare plot, including main plotting area, toolbar, etc...
         either at initialization or global redo'''
        if toolbar:
            plot=ToolbarPlot(self.plotdata, **pltkws)
        else:
            plot=Plot(self.plotdata, **pltkws)
            
 #       plot.active_tool = RangeSelection(plot)
 #       plot.overlays.append(RangeSelectionOverlay(component=plot))    
        
        plot.title = self.plot_title
        plot.padding = 50
        plot.legend.visible=False

        plot.tools.append(PanTool(plot))
        zoom=BetterSelectingZoom(component=plot, tool_mode="box", always_on=False)
        plot.overlays.append(zoom)
        
        #indexlabels=[str(round(i,1)) for i in self.dataframe.index]

        #### If I do plot.index_axis, it actually removes the default values.
    
        #index_axis=LabelAxis(plot, orientation='top', 
                    ##              positions=range(int(float(indexlabels[0])),
                     ##                             int(float(indexlabels[-1]))), 
                                   #positions=range(0,1000,100),
                                #labels=indexlabels,#, resizable='hv',
                                #title=self.indexname)
        #plot.underlays.append(index_axis)

        #plot.value_axis= LabelAxis(plot, orientation='left',  positions=range(self.t_axis_samples),
                                    #mainlabels=['t1', 't2', 't3', 't4','t5', 't6'], resizable='hv', 
                                    #title=self.t_axis_title)
        self.plot=plot
        return

    def _overwrite_plotdata(self):
        '''When a new instance of PandasPlotData is created, this overwrites the
        data source and updates the axis values.'''
        self.plotdata=PandasPlotData(dataframe=self.dataframe)
        #self.plot.index_axis=LabelAxis(self.plot, orientation='bottom', positions=[0,4,42], 
                                #labels=self.plotdata._rowmasks.keys(), resizable='hv',
                                #title=self.indexname)

        #plot.value_axis= LabelAxis(plot, orientation='left',  positions=range(self.t_axis_samples),
                                    #mainlabels=['t1', 't2', 't3', 't4','t5', 't6'], resizable='hv', 
                                    #title=self.t_axis_title)	


    #axis_traits_group=VGroup(
                            #HGroup(
                                    #Item('x_axis_title'), Item('t_axis_title'), Item('plot_title'), 
                                    #),
                            #HGroup(
                                    #Item('markersize'),Item('linestyle'),
                                    #),
                            #)

    #sample_group=Group(
                        #HGroup(Item('spacing'), Item('samp_style') )
                        #)

    main_group=Group(
        Item('plot', editor=ComponentEditor(), show_label=False),
     #   Item('df_new'), Item('df_change'), 
        Item('re_bin', style='simple'),
        #Include('sample_group'),
        #Include('axis_traits_group')
    )

    traits_view=View( Include('main_group') )
    
    

class WithShell(HasTraits):
    df=Instance(DataFrame)
    shell=Any()
    plot=Instance(PandasPlot)
    
    def __init__(self):
        self.df=DataFrame([1,2,3])
        self.plot=PandasPlot()
        self.plot.df=self.df
        
    def _df_changed(self):
        self.plot.df=self.df
        
        
    main_group=Group(
        Item('plot', show_label=False, style='custom'),
     #   Item('df_new'), Item('df_change'), 
        Item('shell', editor=ShellEditor(), style='simple'),
        #Include('sample_group'),
        #Include('axis_traits_group')
    )

    traits_view=View( Include('main_group'), resizable=True)

if __name__=='__main__':
    
  #  df=get_csvdataframe('spectra.pickle')
    df=DataFrame([1,2,3,4,5])


    ### THIS WILL INDUCE A FAILURE IN THE PLOT.PLOT call
 #   df=wavelength_slices(df, ranges=((350.0,370.0), (450.0,500.0), (550.0,570.0), (650.0,680.0), (680.0,700.0)),\
 #                           apply_fcn='simps')

    ### resample dataframe to be a bit less dense, makes plot changes more fluid ###
    
#    df=df.ix[300.0:800.0:5.0, 0:150:5]
 
    theplot=PandasPlot()   
    theplot.df=df   
    theplot.configure_traits()
    
#    test=WithShell()
#    test.df=df
#    test.configure_traits()
    




import numpy as np
from scipy import linalg

def array2d(X, dtype=None, order=None):
    """Returns at least 2-d array with data from X"""
    return np.asarray(np.atleast_2d(X), dtype=dtype, order=order)

def as_float_array(X, copy=True):
    """Converts an array-like to an array of floats

    The new dtype will be np.float32 or np.float64, depending on the original
    type. The function can create a copy or modify the argument depending
    on the argument copy.

    Parameters
    ----------
    X : array

    copy : bool, optional
        If True, a copy of X will be created. If False, a copy may still be
        returned if X's dtype is not a floating point type.

    Returns
    -------
    X : array
        An array of type np.float
    """
    if isinstance(X, np.matrix):
        X = X.A
    elif not isinstance(X, np.ndarray) and not sp.issparse(X):
        return safe_asarray(X, dtype=np.float64)
    if X.dtype in [np.float32, np.float64]:
        return X.copy() if copy else X
    if X.dtype == np.int32:
        X = X.astype(np.float32)
    else:
        X = X.astype(np.float64)
    return X

class PCAError(Exception):
    """ """


# REPLACE V WITH VT TO BE CLOSER IN NOTATION TO NODA BOOK
class PCA():
    """Principal component analysis (PCA)

    Linear dimensionality reduction using Singular Value Decomposition of the
    data and keeping only the most significant singular vectors to project the
    data to a lower dimensional space.

    This implementation uses the scipy.linalg implementation of the singular
    value decomposition. It only works for dense arrays and is not scalable to
    large dimensional data.

    The time complexity of this implementation is ``O(n ** 3)`` assuming
    n ~ n_samples ~ n_features.

    Parameters
    ----------
    n_components : int, None or string
        Number of components to keep.
        if n_components is not set all components are kept::

            n_components == min(n_samples, n_features)

        if n_components == 'mle', Minka\'s MLE is used to guess the dimension
        if ``0 < n_components < 1``, select the number of components such that
        the amount of variance that needs to be explained is greater than the
        percentage specified by n_components

    copy : bool
        If False, data passed to fit are overwritten

    whiten : bool, optional
        When True (False by default) the `components_` vectors are divided
        by n_samples times singular values to ensure uncorrelated outputs
        with unit component-wise variances.

        Whitening will remove some information from the transformed signal
        (the relative variance scales of the components) but can sometime
        improve the predictive accuracy of the downstream estimators by
        making there data respect some hard-wired assumptions.

    Attributes
    ----------
    `components_` : array, [n_components, n_features]
        Components with maximum variance.

    `explained_variance_ratio_` : array, [n_components]
        Percentage of variance explained by each of the selected components. \
        k is not set then all components are stored and the sum of explained \
        variances is equal to 1.0

    Notes
    -----
    For n_components='mle', this class uses the method of `Thomas P. Minka:
    Automatic Choice of Dimensionality for PCA. NIPS 2000: 598-604`

    Due to implementation subtleties of the Singular Value Decomposition (SVD),
    which is used in this implementation, running fit twice on the same matrix
    can lead to principal components with signs flipped (change in direction).
    For this reason, it is important to always use the same estimator object to
    transform data in a consistent fashion.

    Examples
    --------

    >>> import numpy as np
    >>> from sklearn.decomposition import PCA
    >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
    >>> pca = PCA(n_components=2)
    >>> pca.fit(X)
    PCA(copy=True, n_components=2, whiten=False)
    >>> print pca.explained_variance_ratio_ # doctest: +ELLIPSIS
    [ 0.99244...  0.00755...]

    See also
    --------
    ProbabilisticPCA
    RandomizedPCA
    KernelPCA
    SparsePCA
    """
    # Turn off copy?
    def __init__(self, n_components=None, copy=True, whiten=False):
        self.n_components = n_components
        self.copy = copy
        self.whiten = whiten
        
        self._U = None
        self._S = None
        self._VT = None
        
    def fit(self, X, y=None, **params):
        """Fit the model with X.

        Parameters
        ----------
        X: array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        self._fit(X, **params)
        return self

    def fit_transform(self, X, y=None, **params):
        """Fit the model with X and apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        """
        U, S, VT = self._fit(X, **params)
        U = U[:, :self.n_components]

        if self.whiten:
            # X_new = X * V / S * sqrt(n_samples) = U * sqrt(n_samples)
            U *= np.sqrt(X.shape[0])
        else:
            # X_new = X * V = U * S * V^T * V = U * S
            U *= S[:self.n_components]

        return U

    # REPLACE V WITH VT TO BE CLOSER IN NOTATION TO NODA BOOK
    def _fit(self, X):
        X = array2d(X)
        n_samples, n_features = X.shape
        X = as_float_array(X, copy=self.copy)
        # Center data
        self.mean_ = np.mean(X, axis=0) #When transposed, this works fine
        X -= self.mean_
        U, S, VT = linalg.svd(X, full_matrices=False)
        self.explained_variance_ = (S ** 2) / n_samples
        self.explained_variance_ratio_ = self.explained_variance_ / \
                                        self.explained_variance_.sum()

        if self.whiten:
            self.components_ = VT / S[:, np.newaxis] * np.sqrt(n_samples)
        else:
            self.components_ = VT

        if self.n_components == 'mle':
            if n_samples < n_features:
                raise ValueError("n_components='mle' is only supported "
                "if n_samples >= n_features")
            self.n_components = _infer_dimension_(self.explained_variance_,
                                            n_samples, n_features)

        elif (self.n_components is not None
              and 0 < self.n_components
              and self.n_components < 1.0):
            # number of components for which the cumulated explained variance
            # percentage is superior to the desired threshold
            ratio_cumsum = self.explained_variance_ratio_.cumsum()
            self.n_components = np.sum(ratio_cumsum < self.n_components) + 1

        if self.n_components is not None:
            self.components_ = self.components_[:self.n_components, :]
            self.explained_variance_ = \
                    self.explained_variance_[:self.n_components]
            self.explained_variance_ratio_ = \
                    self.explained_variance_ratio_[:self.n_components]

        self._U = U
        self._S = S
        self._VT = VT
        
        return (U, S, VT)
    


    @property
    def U(self):
        """ Score matrix (in Noda = W) """
        if self._U is None:
            raise PCAError("Please run .fit() to set U, S, VT from SVD.")
        return self._U
        

    @property
    def S(self):
        """ """
        if self._S is None:
            raise PCAError("Please run .fit() to set U, S, VT from SVD.")       
        return self._S        
        
    @property
    def VT(self):
        """ Loading Vectors """        
        if self._VT is None:
            raise PCAError("Please run .fit() to set U, S, VT from SVD.")
        return self._VT
    
    @property
    def W(self):
        """ Score matrix"""
        return self.U * self.S #ok?

    def transform(self, X):
        """Apply the dimensionality reduction on X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            New data, where n_samples in the number of samples
            and n_features is the number of features.

        Returns
        -------
        X_new : array-like, shape (n_samples, n_components)

        """
        X_transformed = X - self.mean_
        X_transformed = np.dot(X_transformed, self.components_.T)
        return X_transformed

    def inverse_transform(self, X):
        """Transform data back to its original space, i.e.,
        return an input X_original whose transform would be X

        Parameters
        ----------
        X : array-like, shape (n_samples, n_components)
            New data, where n_samples in the number of samples
            and n_components is the number of components.

        Returns
        -------
        X_original array-like, shape (n_samples, n_features)

        Notes
        -----
        If whitening is enabled, inverse_transform does not compute the
        exact inverse operation as transform.
        """
        return np.dot(X, self.components_) + self.mean_
    

    #def __repr__(self): 
        #return 'YEAHH BOY'


''' Series of utilities for determining the size of spherical gold nanoparticles 
    from the optical aborsbance spectra, as published in:

    Haiss, Thanh, Aveyard, Fernig, Determination of Size and Concentration of Gold 
    Nanoparticles from UV-Vis Spectra Anal. Chem, 79, 2007.

    Three functions are build based on the results of this paper, as well as a utility
    for estimating nanoparticle concentration.'''

import math
import numpy as np
import warnings

def _haiss_preformat(ts, style='boxcar', width=None, limit_range=(400.0,700.0)):
    '''Called by other haiss method to preforma the timespectra before calling
       haiss functions.
       
        See description under haiss_m1 for explanation of parameters.  '''


    ### Ensure ts is in units of absorbance and has nanometer spectral unit
    if ts.iunit != 'a':
        try:
            ts.iunit='a'
        except Exception:
            raise TypeError('%s must be in absorbance units for quick_haiss_m1 to function.'%ts.name)
        else:
            warnings.warn('Converted iunit to absorbance in haiss function.')
        
    
    if ts.specunit != 'nm':
        try:
            ts.specunit='nm'
        except Exception:
            raise TypeError('%s must have specunits of nm to be consistent with published formulae.'%ts.name)
        else:
            warnings.warn('Converted specunit to nm in haiss function.')
            
                    
  
    if width==0.0:
        width=None

    ### Smoothing    
    if width:
        if style=='boxcar':
            try:
                ts=ts.boxcar(width)
            except AttributeError:
                raise AttributeError('Cannot find boxcar() method on %s object.'%ts.name)
        else:
            raise NotImplementedError('Only boxcar smoothing is available')
    
    ### Slice data
    if limit_range:        
        if len(limit_range) != 2:
            raise AttributeError('Limit_range must be a length-2 iterable.')
        
        ts=ts.ix[ limit_range[0]:limit_range[1] ]    
        
    return ts

def haiss_m1(ts, style='boxcar', width=None, limit_range=(400.0,700.0)):
    ''' Wrapper function for Haiss method  1.  Allows for range slicing/smoothing and applies
        haiss_m1 over the entire timespectra object.

        Parameters:
        ---------
  
        ts: TimeSpectra.
        style: Smoothing style.  For now, only boxcar is used.
        width: Smoothing binwidth in units of ts.specunit.  If None or 0, no smoothing is performed.
        limit_range: width of the range used (default is (400.0,700.0)
    
        Notes:
        -----------
    
        TimeSpectra must have iunit of absorbance.  Code attempts to convert it but does not attempt
        to add a reference if ts.iunit results in an error.  
        
        Boxcar is an INSTANCE METHOD of TimeSpectra.  A boxcar function that returns a dataframe can 
        be found in skspec.core.utilities.
        
        Code may error if user passes a curve with
        a maximum very far from 510 (lam0 in haiss_m1) as the log will blowup.         
    
    '''

    ts=_haiss_preformat(ts, style=style, width=width, limit_range=limit_range)

    ### Feed index at maxiumum. to haiss function
    spr=ts.idxmax()    
    return spr.apply(_haiss_m1)
        

def haiss_m2(ts, ref=450.0, width=None, style='boxcar', ref_width=None, 
                    peak_width=None, limit_range=(400.0,700.0), exp=True):
    ''' Wrapper for haiss_m2 to return diamter of nanoparticles based on features of its
        absorbance spectrum relative to water.  Program has special call options to handle
        smoothing and range slicing on the fly.

    Parameters:
    ----------
    curve: 
    smooth: If True, applies smoothing to ts.
    boxcar: The value of boxcar when smoothing the spectrum.
    width: Smoothing binwidth in units of ts.specunit.
    limit_range: width of the range used (default is (400.0,700.0)
    peak_width: +/- wavelength to average around peak value when determining Absorbance at SPR.
    Xspr: x value (wavelength) corresponding to max y
    Aspr: if using peak_width, uses average value of absorbance in range [Xspr-peak_width : Xspr+peak_width] as maximum absorbance. Otherwise, defaults to max value.
    ref_width: The width of the range designated for reference point window
    Aref: if using ref_width, uses average value of absorbance in range [refspr-ref_width : refspr+ref_width] as reference value. Otherwise, defaults to max value
    exp: If true, use experimental parameters from Haiss et. al. to do fit; otherwise use theoretical.

    Notes:
    -----------
    TimeSpectra must have iunit of absorbance.  Code attempts to convert it but does not attempt
        to add a reference if ts.iunit results in an error.  
        
        Boxcar is an INSTANCE METHOD of TimeSpectra.  A boxcar function that returns a dataframe can 
        be found in skspec.core.utilities.

        Code may error if user passes a curve with
        a maximum very far from 510 (lam0 in haiss_m1) as the log will blowup.
    '''
    
    ts=_haiss_preformat(ts, style=style, width=width, limit_range=limit_range)
    
    return ts.apply(_m2_haiss, ref=ref, peak_width=peak_width, ref_width=ref_width, exp=exp)


def _m2_haiss(curve, ref=450.0, peak_width=None, ref_width=None, exp=True):
    ''' Private method to quick_haiss_m2.  This separation is necessary to allow
        haissm2 (with 2 required arguments) to be compatable with quick_haiss_m2.
        In other words, because we need to compute Aspr and Aref separately, can't
        just call .apply() method in the normal fashion as done for example
        in _haiss_m1/_haiss_m3. This handles the single curve/series operations.
    '''

	
    ### Get Aspr from peak value with or w/o averaging neighbors    
    if peak_width:
        Xspr=curve.idxmax(axis=0)
        Aspr=curve[Xspr-peak_width : Xspr+peak_width].mean() 
    else:
        Aspr=curve.max(axis=0)

    ### Get Aref from peak value with or w/o averaging neighbors
    if ref_width:
        Aref=curve[ref-ref_width : ref+ref_width].mean()
    else:
        Aref=curve.ix[ref]   #FIND CLOSEST WAVELENGTH
                
 
    if Aspr==0.0 or Aref==0.0:
        d=np.NaN
    else:
        d= _haiss_m2(Aspr, Aref, exp=exp)
    return d

def _haiss_m1(lambda_spr):
    ''' Return diameter from wavelength at spr maximum of absorbance curve.

    Parameters:
    -----------
    lambda_spr: the wavelength at spr absorbance maximum.

    Notes:
    -----------
    this method is not accurate for particles with diameter less than 25nm.

    '''
    lambda_0=512
    L1=6.53
    L2=0.0216

    d=(math.log((lambda_spr-lambda_0)/L1))/L2
    return d


def _haiss_m2(Aspr, A450, exp=True):
    ''' Return diameter from absorbance at spr and 450 from Haiss et al. 

    Parameters:
        -----------
           Aspr: the absorbance at spr max.
           A450: the absorbance at 450nm.
           exp: If true, experimentally derived Haiss fit parameters are used.
                Otherwise, theoretical Haiss fit parameters are used.

        Notes:
        -----------
           the diameter is calcualted by dividing the absorbance at the spr 
           maximum by that at 450nm and fitting to experimental/theoretical
           parameters.
    '''

    ### Calculate diameter from experimental fit parameters
    if exp:
        B1=3.00 
        B2=2.20
        d=math.exp( (B1 * (Aspr/A450) - B2) )

    ### Calculate diameter from theoretical fit parameters 
    else:
        B1=3.55
        B2=3.11
        d=math.exp( (B1 * (Aspr/A450) - B2) )

    return d

def haiss_m3(ts, Cau, dilution=None, style='boxcar', width=None, limit_range=(400.0,700.0), peak_width=None, exp=True):
    ''' Wrapper function for Haiss method  3.  Estimates diamter of gold nanoparticles
        based on the SPR max of the absorbance and the initial concentration of gold use to synthesize
        the nanoparticles.  Should be valid between 5-50nm diatmeters.  Please take note of the dilution
        parameter description below.

        Parameters:
        ---------
  
        ts: TimeSpectra.
        Cau: Initial concentration of gold used to synthesize nps (in mol/L)
        dilution: This method requires absorbance of FULLY concentrated np solution.  If data was taken for
                  a dilute conenctration, say 10% stock NP concentration, then use dilution=0.1.
        style: Smoothing style.  For now, only boxcar is used.
        width: Smoothing binwidth in units of ts.specunit.  If None or 0, no smoothing is performed.
        limit_range: width of the range used (default is (400.0,700.0)
        peak_width: +/- wavelength to average around peak value when determining Absorbance at SPR.
        exp: If true, use experimental parameters from Haiss et. al. to do fit; otherwise use theoretical.

    
        Notes:
        -----------
    
        TimeSpectra must have iunit of absorbance.  Code attempts to convert it but does not attempt
        to add a reference if ts.iunit results in an error.  
        
        Boxcar is an INSTANCE METHOD of TimeSpectra.  A boxcar function that returns a dataframe can 
        be found in skspec.core.utilities.
        
        We've independently verified that Absorbance at relevant wavelengths (400-700nm) scales linearly 
        with concentration from dilutions as low as 1% up to fully concentrated NP's.  By fully concentrated,
        we mean the typical batch concentration made from the Turkevich protocol (Generally 10^12-14 np's/mL)
        In extremely dilute or concentrate solutions, this linearity may breakdown and the accuracy of this 
        function is reduced.

    '''

    ts=_haiss_preformat(ts, style=style, width=width, limit_range=limit_range)

    ### Feed index at maxiumum. to haiss function
    ### Get Aspr from peak value with or w/o averaging neighbors    
    if peak_width:
        ### Need to compute a slice average based on SPR xvalue (bit tricky when ts not a single curve)
        Xspr=ts.idxmax(axis=0)
        xrng0=Xspr-peak_width
        xrng1=Xspr+peak_width
        xavg=(xrng0 + xrng1) / 2.0
        Aspr=ts.ix[xavg].mean()
    else:
        Aspr=ts.max(axis=0)
        
    ### Scale up Aspr based on dilution factor.
    if dilution:
        Aspr=Aspr/dilution

    return Aspr.apply(_haiss_m3, Cau=Cau, exp=exp)


def _haiss_m3(Aspr, Cau, exp=True):
    ''' See descriptoin for haiss_m3 '''

    term=5.89*10**-6

    ### Calculate diameter from experimental fit parameters
    if exp:
        C1=-4.75
        C2=0.314 

    ### Calculate diameter from theoretical fit parameters 
    else:
        C1=-4.70
        C2=0.300

    return ( (Aspr * term )/ (Cau * math.exp(C1) ) )**(1.0/C2)

def haiss_conc(ts, d, style='boxcar', width=None, limit_range=(400.0,700.0), ref=450.0, ref_width=None, exp=True):
    ''' Return estimation of AuNP concentration given the diameter of gold nanoparticles as the Absorbance
        at 450nm (or other wavelength controlled by ref).  Should be valid for particles with diameters ranging
        from 2.5nm to 100nm and possibly beyond these limits.  Works for stock and dilute samples.
        
        Parameters:
        ---------
  
        ts: TimeSpectra.
        d: Size of gold nanoparticles in nm
        style: Smoothing style.  For now, only boxcar is used.
        width: Smoothing binwidth in units of ts.specunit.  If None or 0, no smoothing is performed.
        limit_range: width of the range used (default is (400.0,700.0)
        ref: Wavelength used for calibration in haiss paper (default is 450nm)
        ref_width: +/- wavelength to average around peak value when determining Absorbance at ref.
        exp: If true, use experimental parameters from Haiss et. al. to do fit; otherwise use theoretical.
       
        Notes:
        -----
          I verified that dilution is not a factor by including it into some trials.  Since N scales linearly 
          with A450, then a dilution of .1 yeilds 1/10th N and so forth.  Therefore, uses can take results form
          this function, and scale them up by the dilution factor to get the full-batch number density.
        
        '''
    ts=_haiss_preformat(ts, style=style, width=width, limit_range=limit_range)
    ### Get Aref from peak value with or w/o averaging neighbors
    if ref_width:
        Aref=ts.ix[ref-ref_width : ref+ref_width]  #Stil slices by wavelength
    else:
        Aref=ts.ix[ref]  #FIND CLOSEST WAVELENGTH    
    
    return Aref.apply(_haiss_conc, d=d)


def _haiss_conc(Aref, d):
    ''' See haiss conc '''
    num=Aref * 10**14
    t1=1.36 * math.exp(-( (d-96.8)/(78.2))**2 )
    den=d**2 * (-0.295+ t1)
    return num/den



import sys
import os
import os.path as op
from collections import OrderedDict

from skspec.scripts.gwu_script.tex_templates import EXPERIMENT_MAIN, HEADER, \
    BLANK_SECTION

import logging
logger = logging.getLogger(__name__)
from skspec.logger import log, configure_logger, logclass

from skspec.scripts.gwu_script import LOGO_PATH

logger = configure_logger(name=__name__)

@logclass(public_lvl='info')
class Reporter(object):
    ''' Report is stored as an ordered dict.  Main is followed by sections. '''
    
    def __init__(self, **kwargs):
       
        self.overwrite = kwargs.get('overwrite', False)
       
        # Add main experimental body from template file or blank
        template_file = kwargs.get('template_file', None)
        if template_file:
            if not op.exists(template_file):
                raise IOError("Template file %s not found!")
            self.main_template = file(template_file, 'r').read() 

        else:
            self.main_template = EXPERIMENT_MAIN
        
        self.title = kwargs.get('title', 'Untitled')
        self._author = kwargs.get('author', 'Adam Hughes')
        self.email  = kwargs.get('email', 'hugadams@gwmail.gwu.edu')

        # Handle sections
        self.sections = OrderedDict()
        self.parse_sections(kwargs.get('sections', []))

    @property
    def author(self):
        ''' (First, Last) name of author '''
        try:
            # BUG: This will put > 2 arguments into last name.  IE Adam STEPHEN HUGHES
            first, last = self._author.split(' ', 1)
            return (first, last)
        except Exception:
            raise AttributeError('Author must first and last name of the form:' 
                                 ' First Last')

    def parse_path(self, outpath):
        ''' Handles path overwriting of a single file. '''

        outpath = op.abspath(outpath)
        
        if op.exists(outpath):
            if not self.overwrite:
                # Replace these IOErrors everywhere with PathError?
                raise IOError('Outpath already exists: "%s"', outpath)
            logger.warn('Overwriting outpath: "%s"' % outpath)
            os.remove(outpath)

        return outpath
    
    def parse_sections(self, sections):
        ''' Adds sections to self.sections.  If length 1 iterable and the 
           entry is a file, treats this a a tree files and reads from tree file.
           Otherwise, adds blank sections.'''

        if not hasattr(sections, '__iter__'):
            raise AttributeError('Reporter.parse_sections must be iterable.')
        
        # Assumes section is a tree file
        if len(sections) == 1 and op.exists(sections[0]):
            logger.info("Trying to parse sections from treefile!")
            sections_dict = eval(open(sections[0], 'r').read())
            
            for secname, section_file in sections_dict.items():
                self.add_section(secname, section_file)

        # Adds a blank section for each name in order as (section, SEC_TEMPLATE)
        else:
            for secname in sections:
                self.add_section(secname)
                
    @property
    def report(self):
        ''' Concatenates main_tamplate body and self.sections. '''
        body = self.main_template
        for secname, section in self.sections.items():
            body += section + '\n'
        return body   
            
        
    def add_section(self, secname, section_file=None):
        ''' If not section_file, blank template is used. '''

    #    secname = self.latex_secname(secname)

        if section_file:
            logger.info("Appending section string from file file for %s" % secname)
            self.sections[secname] =str(open(section_file, 'r').read())

        else:
            # Need to fill in "secname" via %, while treefile already did this
            logger.info("Appending in blank section %s" % secname)
            self.sections[secname] = BLANK_SECTION  % {'secname':secname}
        
        
    def remove_section(self, secname):
        try:
            del self.sections[secname]
        except KeyError:
            raise KeyError('%s not found in "sections".  See self.sections')


    def output_template(self, outpath, add_header=False):

        outpath = self.parse_path(outpath)
        open(outpath, 'w').write(self.report)
        logger.info('Template written to: "%s"' % outpath)

                                                        
    def make_report(self, template_file, outpath):
        ''' Joins all the section strings from self.report.values();
            adds a header and mergers tex_params (author, email etc..).'''
        
                        
        if not op.exists(template_file):
            raise IOError('Template file not found: "%s"' % template_file)
        
        # Add the full report as the "body" into the header template_file
        tex_params = {}
        tex_params['title'] = self.title
        tex_params['logopath'] = LOGO_PATH
        tex_params['authorfirstname'] = self.author[0]
        tex_params['authorlastname'] = self.author[1]
        tex_params['email'] = self.email
        tex_params['body'] = open(template_file, 'r').read()
        
        outpath = self.parse_path(outpath)       
        open(outpath, 'w').write(HEADER % tex_params) 
        logger.info('Report written to: "%s"' % outpath)



from pandas import Float64Index, Index
import numpy as np
from skspec.units.abcunits import UnitError, Unit, ConversionUnit


# Parsing Functions (because need access before self exists)
# ----------------------------------------------------------
def _parse_unit(unit):
   if unit is not None:
      if not isinstance(unit, Unit):
         raise UnitError('Expected units.Unit type, got %s' \
                     % (type(unit)))
   return unit      


def _parse_conversion_unit(unit, unitdict):
   """ Given a string unit (ie nm), returns the corresponding unit
   class."""
   if unit not in unitdict:
      raise UnitError('Invalid unit "%s".  Choose from %s' % 
                      (unit, sorted(unitdict.keys()) ) )
   return unitdict[unit]  


# Custom Index Classes
# --------------------

# NOT SURE WHY BUT SUBCLASSING INDEX CALLS THINGS LIKE ALL_DATES THAT SCREW THIS UP
# IN CERTAIN CASES BUT THIS ACTUALLY WORKS, EVEN WHEN DTYPE IS TIMESTAMPS!?
class CustomIndex(Float64Index):
   """ Custom index used in skspec to interface to Unit."""
   
   _unit = None #leave as _unit to keep api of Spectra

 #  def is_all_dates(self):
 #     return True
   
#   def is_datetime_array(self):
#      return True

   def __array_finalize__(self, obj):
      """ Used in slicing and other places, not sure exactly.  Obj is parent
      object, transferring attributes."""
      if obj is None:
         return 
      
      self._unit = getattr(obj, '_unit', None) 
      self._id = getattr(obj, '_id', None) #Used by pandas object in copying I think
      
   def __new__(cls, input_array, unit=None):
      """ Unit is valid key of unitdict """
      
      # All that's needed, yes?
      obj = np.asarray(input_array).view(cls)                 
      obj._unit = _parse_unit(unit)
      return obj
   
   #def __getattr__(self, attr):
      #""" Defer attribute call to self._unit"""
      #try:
         #return getattr(self._unit, attr)
      #except AttributeError:
         #raise AttributeError('%s has no attribute "%s".' % 
                              #(self.__class__.__name__, attr) )

   def __unicode__(self):
      """ Returned on printout call.  Not sure why repr not called... """
      out = super(CustomIndex, self).__unicode__()        
      return out.replace(self.__class__.__name__, '%s[%s]' % 
                         (self.__class__.__name__, self._unit.short))
   
   def convert(self, outunit):
      """ """
      if outunit is not None:
         if not isinstance(outunit, Unit):
            raise UnitError("%s expects skspec.unit.Unit, got %s.  Were you" \
                        " expecting a ConversionIndex?" \
                        % (self.__class__.__name__, type(outunit)))
                        
      self._unit = outunit   

   @property
   def unit(self):
      return self._unit


class ConversionIndex(Index):
   """ Base class for skspec.  To overwrite, requires:
        - replace unitdict with a dictionary from units.py, eg SPECUNITS
        - overwrite convert(), which defines conversions on the unit types.
   Since we are subclassing a numpy array and not a basic python object, 
   we don't write __init__(); rather, __new__ and __array_finalize__ are
   employed (http://docs.scipy.org/doc/numpy/user/basics.subclassing.html)
   
   _forcetype will force any construction to this type.  Even tho it inherits
   from float64index, found that passing an Int64Index to SpecIndex() resulted
   in integer indicies.  
   """

   unitdict = None 
   addnullunit = True
   _forcetype = None 
   

   def __new__(cls, input_array, unit=None):
      """ Unit is valid key of unitdict """
      #Not sure why I have to enfory dtype, but if SpecIndex() of an
      # Int64 index, it remains integer
      if cls._forcetype:
         obj = np.asarray(input_array, dtype=cls._forcetype).view(cls)
      else:
         obj = np.asarray(input_array).view(cls)         
      obj._unit = _parse_conversion_unit(unit, cls.unitdict)
      return obj

   # I am not really worred about other constructors yet...
   def __array_finalize__(self, obj):
      """No matter what constructor called, this will get called, so does
      all the housekeeping."""

      if None not in self.unitdict:
         self.unitdict[None] = Unit()
      
      if obj is None: 
         return
      
      # Can be called form slicing or other reasons, and I don't always know
      # if unit is still a string/Non_e, or if it's already a Unit class.
      # Thus, I let both cases work, even though don't know when is what
      unit = getattr(obj, 'unit', None)
      self._id = getattr(obj, '_id', None)
      
      # Necessary for certain instantiations DONT CHANGE
      if isinstance(unit, basestring) or unit is None:
         self._unit = _parse_conversion_unit(unit, self.unitdict)
      else:
         self._unit = unit

   def convert(self, outunit):
      """Convert spectral values based on string outunit.  First converts
      the current unit to the canonical unit (eg, nanometers goes to meters)
      then to the outunit (eg, meters to eV).  This is done through the unit
      methods, .to_canonical() and .from_canonical() where in the case of
      spectral units, canonical refers to meters.
      """
      outunit = _parse_conversion_unit(outunit, self.unitdict)
      inunit = self._unit

      # Unit not changed, or set to None
      if outunit.short == inunit.short or not outunit.short:
         return self.__class__(self, unit=outunit.short)

      # If current unit is None, just set new unit
      if not inunit.short:
         return self.__class__(self, unit = outunit.short)

      # Convert non-null unit to another non-null unit   
      else: 
         canonical = inunit.to_canonical(np.array(self))
         arrayout = outunit.from_canonical(canonical)
         return self.__class__(arrayout, unit=outunit.short)      
      

   #Email list about this distinction
   #def __repr__(self):
      #""" Used internally in code, like if I print self fromin function."""
      #out = super(ConversionIndex, self).__repr__()        
      #return out.replace(self.__class__.__name__, '%s[%s]' %
                         #(self.__class__.__name__, self._unit.short))

   def __unicode__(self):
      """ Returned on printout call.  Not sure why repr not called... """
      out = super(ConversionIndex, self).__unicode__()        
      return out.replace(self.__class__.__name__, '%s[%s]' % 
                         (self.__class__.__name__, self._unit.short))

   # PROMOTED UNIT METHODS
   #def __getattr__(self, attr):
      #""" Defer attribute call to self._unit"""
      #try:
         #return getattr(self._unit, attr)
      #except AttributeError:
         #raise AttributeError('%s has no attribute "%s".' % 
                              #(self.__class__.__name__, attr) )

   @property
   def unit(self):
      return self._unit.short

   @property
   def unitshortdict(self):
      """ Return key:shortname; used by TimeSpectra to list output units."""
      return dict((k,v.full) for k,v in self.unitdict.items())
      
   
class ConversionFloat64Index(ConversionIndex, Float64Index):
   """ Float64 conversion index.  Main difference is that Float64Index
   has several properties like "is_all_dates" and "is_unique" that are 
   critical for slicing and other Dataframe operations.  Therefore, we 
   want to retain all these for the Float64Index.  By method res order (MRO),
   they should not overwrite anything define don conversion index.  
   
   More about MRO in python: http://mypythonnotes.wordpress.com/2008/11/01/
                 python-multiple-inheritance-and-the-diamond-problem/ """
   
   _forcetype = 'float64' 
   
   #def __new__(cls, input_array, unit=None):
      #""" Unit is valid key of unitdict """

      #input_array = np.around(input_array, decimals=6)
      #return super(ConversionFloat64Index, cls).__new__(cls, input_array, unit=unit)
   
   


#CITE BOOK, REWRITE

""" Suite for 2-dimensional correlation analysis.  Decided to put it in, despite already
seeing a numpy-based version in paper "2D-correlation analysis applied to in situ and operando Mossbauer spectroscopy".
This would not be pandas friendly, so it wouldn't intgerate into skspec and I also want to build some tools from 
scratch to refamiliarize myself with the topic.

To remember/writeup eventually:

From page 32 Noda book,
Classical statistical cross correlation measures the dot product between spectral variables at different times.
For example, a pulse at t=0 and again at t=3, then the cross correlation would be max at tau=3.  If this experiment
went on for 50 minutes, then we average and integrate over all time.  If this pulse happed every 3 seconds, this correlation
would be 100% for exmaple.  If the pulse happed like only 5 times, the correlation would be diluted by the normalization factor.

The synchronous spectrum is the cross correlation at tau=0, summed and averaged over all times.  Basically, it measures the instantaneous
correlation of all spectral variables.  The asynchronous spectrum is the instantaneous orthogonal spectrum basically.  So whenever two things
are moving together at one instant, weight goes into the sync.  If not, weight goes into the async  Makes me think of an easy filter!  Can
just throw out regions where not much is happening right?  Maybe call it a 2dcs filter.


2d Codistriubtion Spectroscopy[2] is a more recent development and is also included.

References
----------

[1] Scaling techniques to enhance two-dimensional correlation spectra, Isao Noda,
    Journal of Molecular Structure.  2008. Volume: 883-884, Issue: 1-3, (216-227) 
[2] Isao Nodal. Journal of Molectul Structure. 2014.  Two-dimensional codistriubtion spectroscopy to
    determine the sequential order of distributed presence of species.  1069. 2014 (50-59)

"""

#Copy these to notebooks?

from __future__ import division

import logging
import pandas
logger = logging.getLogger(__name__) 

from math import pi
import numpy as np

from skspec.core.anyspectra import AnyFrame 
from skspec.plotting.correlation_plot import corr2d, corr3d, corr_multi
import skspec.config as pvconfig
import skspec.core.utilities as pvutils
from skspec.pandas_utils.metadframe import MetaDataFrame
from pca_lite import PCA
#from pcakernel import PCA


# Original way to subtract ref spec:
#     data_trans = matrix.transpose()
#    return (data_trans - vector).transpose()

def noda_matrix(length):
    ''' Length is the number of timepoints/columns in the dataframe. 
       Returns the hilbert noda Transformation matrix.'''

    # XXX: how to vectorize?
    Njk = np.empty( (length,length))
    for j in range(length):
        for k in range(length):
            if j==k:
                Njk[j,k]=0
            else:
                Njk[j, k]=1.0 /  ( pi * (k-j)  )  #DOUBLE CHECK j-i with old verison.    
    return Njk


class CorrError(Exception):
    """ """
    
class Spec2dError(Exception):
    """ """

class Spec2d(AnyFrame):
    """ Frame to hold synchronous, asynchronous and other 2D spectra objects.
    Mostly just want to customize headers, and subclasses correspond to various
    corr2D object.  For example, synchronous spectra has spectral data on 
    index and columns.  Therefore, it is represented by Spec2D class.

    Also overwrites Spectra.norm for 3dPlotting purposes.

    Will update active/spectral unit based on index/columns, but original
    datarange (like how much time synchronous corresponds to) is stored
    permanently in self._span_string
    """

    def __init__(self, *args, **kwargs):
        """ Hack over norms for plotting API sake.  Stores metadata like
        scaling, spectral and variance range of spectra.  *args, **kwargs
        passed to AnyFrame constructor.

        kwargs
        ------

        _corr2d: Corr2D object

        _norm: ''
           norm to appear in 3dplots and on contour colorbar
        """


        # Mandatory, access to corr2d and hence original dataset 
        # (passed by reference)!
        self._corr2d = kwargs.pop('corr2d')

        #self._scalestring = kwargs.pop('scalestring', '??')
        #self._originadataindex

        # INDEX/COLUMNS are SET TO INDEX OF ORIGINAL DATA
        kwargs['index'] = self._corr2d.index
        kwargs['columns'] = self._corr2d.index

        super(Spec2d, self).__init__(*args, **kwargs)


    @classmethod
    def from_corr2d(cls, corr2d):
        return cls(scale_string=corr2d.scalestring, 
                   originaldatindex=corr2d.index)
        
    @property
    def _var_span(self):
        cols = self._corr2d.columns
        return pvutils._compute_span(cols, with_unit=True)
    
    # Header/output
    # -------------
    @property
    def _header(self):
        """ Header for string printout """
        delim = pvconfig.HEADERDELIM

        # Since correlation REQUIRES index units, this will probably just work
        
        # Certain methods aren't copying this correctly.  Delete later if fix
        name = pvutils.safe_lookup(self, 'name')
        ftunit = pvutils.safe_lookup(self, 'varunit')  
        spunit = pvutils.safe_lookup(self, 'specunit')


        header = "*%s*%sScaling:%s\n" % \
            (name, delim, self._corr2d._scale_string)

        return header


    @property
    def _header_html(self):
        """ Header for html printout """
        delim = pvconfig.HEADERHTMLDELIM

        # Should be same by virtue of index units being same, but don't
        # enforece it; up to user not to change it!
        ftunit = pvutils.safe_lookup(self, 'varunit')  
        spunit = pvutils.safe_lookup(self, 'specunit')

        if self.ndim > 1:
            s1, s2 = self.shape            
            colorshape = '<font color="#0000CD">(%s%s X %s%s)</font>' % (
                s1,
                spunit,
                s2,
                ftunit)

        #If user turns 2D Corr into a 1D spectrum (which they shouldn't...)
        else:
            colorshape = '<font color="#0000CD"> (%s)</font>' % (self.shape)

        # Green        
        scale_string = 'Scaling: <font color="#197519">%s</font>' % self._corr2d._scale_string

        # Black 
        span_string = 'Span:  <font color="#000000">%s</font>' % self._var_span

        header = "%s:&nbsp%s%s%s%s%s\n" % \
            (self.name, 
             colorshape,
             delim,
             scale_string,
             delim,
             span_string         
             )   

        return header


    def plot(self, kind='corr2d', **pltkwargs):
        """ """

        pltkwargs.setdefault('title', '%s (%s)' % (self.name, self._var_span))        

        # Default to specplot
        if kind not in ['corr2d', 'corr3d']:
            return super(Spec2d, self).plot(kind=kind, **pltkwargs)

        pltkwargs.setdefault('xlabel', self.specunit)
        pltkwargs.setdefault('ylabel', self.specunit)         

        if kind == 'corr2d':
            # If user sets color/cmap, will overwrite
            if 'cmap' not in pltkwargs and 'color' not in pltkwargs and 'colors' not in pltkwargs:
                pltkwargs['cmap'] = pvconfig.CMAP_CONTOUR

            if self.index[0] > self.index[-1]:
                pltkwargs['_reverse_axis'] = True


            return corr2d(self, **pltkwargs)

        elif kind == 'corr3d':
            return corr3d(self, **pltkwargs)




# Keep this independt of TS; just numpy then more flexible
class Corr2d(object):
    """ Computed 2d correlation spectra, including synchronous and asynchronus,
    correlation, disrelation and other spectra given a 2d data matrix, index
    and columns.  Index and columns are necessary for plotting, so made them
    a mandatory requirement."""

    # Columns aren't used; should I eliminate
    def __init__(self, spec, refspec=None):
        """ refspec is if you want custom centering.  """
        if spec.ndim != 2:
            raise CorrError('Data must be 2d!')

        if not isinstance(spec, MetaDataFrame):
            raise CorrError('Corr2d requires skspec data structures (Metadataframe,'
                            'Spectra, etc... got %s') % type(data)


        # MAKE AN ACTUAL COPY OF DATA, NOT PASSING BY REFERENCE
        self.spec = spec.deepcopy()


        self.index = spec.index   #Relax these maybe and just hide some sideplots...
        self.columns = spec.columns
        self.specunit = spec.specunit
        self.varunit = spec.varunit

        # Better to store than compute as a property over and over
        self._noda = noda_matrix(self.M)

        # Defaults
        self._scaled = False
        self.alpha = 0.8
        self.beta = 0.0
        self._PCA = None

        # Ref spectrum/dynamic spectrum/centering
        if refspec is not None:
            # QUICKEST VAILDATION OF REF_SPEC (WHY CLASS NOT WORKING WIT NP.NDARRAY)
            
            #INSTEAD OF TYPE CHECK, JUST FORCE CONVERT BY DOING array(REFSPEC)
            refspec = np.array(refspec)
            if refspec.shape != self.index.shape:
                raise CorrError('Shape mismatch: spectral data index %s and'
                                ' reference spec shape %.' % \
                                (self.index.shape, refspec.shape))

            # Ref spectrum must be stored as an array for subtraction to work 
            # as defined here!
            self.ref_spectrum = np.array(refspec)           
            self._center = 'Pre-centered'
            self.dyn_spec = self.spec.subtract(self.ref_spectrum, axis=0)

        else:
            self.set_center('mean')


    # Scaling and Centering
    # ---------------------

    @property
    def _scale_string(self):
        """ Current state of scalling, used in __repr__ and plot """
        if self._scaled:
            return '(a=%s, b=%s)' % (self.alpha, self.beta)
        return 'False'


    def scale(self, *args, **kwargs):
        """Scale the synchronous and asynchronous spectra via REF 2
        based on generalized exponential parameters.  Scaling alpha will enhance
        the fine detail of the correlations, but also the noise.  Enhancing beta
        can screen the fine details and enhance the primary correlations.  Alpha
        0.8 and beta 0.0 are suggested as optimal tradeoff between fine correlation
        enhancement and low noise.
        """
        if args:
            if len(args) > 1:
                raise CorrError('Please use keywords (alpha=..., beta=...) to'
                                'avoid ambiguity.') 
                # Make that a custom exception
            if args[0] == True:
                if self._scaled == True:
                    logger.warn("Data already scaled!")
                self._scaled = True
            elif args[0] == False:
                if self._scaled == False:
                    logger.warn("Data already unscaled!")
                self._scaled = False
            else:
                raise CorrError('Argument "%s" not understood; please use True or False.')

        else:
            self._scaled = True

        self.alpha = kwargs.pop('alpha', self.alpha)
        self.beta = kwargs.pop('beta', self.beta)
        if self.alpha > 1 or self.beta > 1:
            logger.warn('Alpha/Beta lose meaning off of range 0-1.')


    @property
    def center(self):
        return self._center
    

    def set_center(self, style, *args, **kwargs):
        """ User sets centering, this updates the  """

        try:

            if not style:
                self._center = None
                # Instead of np.zeros, I will just multiply 0 times a column
                # To get spectrum of 0's
                ref_spectrum = self.spec[0] * 0.0  

            elif style == 'mean':
                self._center = 'mean'
                ref_spectrum = self.spec.mean(axis=1)

            # style is understood as a function, but not inspected
            else:            
                self._center = 'custom fcn.'           
                ref_spectrum = style(self.spec, *args, **kwargs)

        except Exception:

            raise CorrError('Center requires style of "mean", None or a '
                            ' a function, got "%s".' % style)

        self.ref_spectrum = np.array(ref_spectrum)

        if len(self.ref_spectrum) != self.shape[0]:
            raise CorrError('ref. spectrum should be of spectral length (%s)'
                            ' got "%s".' % (self.shape[0], len(self.ref_spectrum)))

        # Set dynamic spectrum.  Should just be able to subtract but numpy messing up        
        self.dyn_spec = self.spec.subtract(self.ref_spectrum, axis=0)


    @property
    def shape(self):
        return self.spec.shape     

    @property
    def M(self):
        return self.shape[1]

    # Numpy Arrays
    # ------------

    @property
    def sync_noscale(self):
        """ Return unscaled, synchronous spectrum as a numpy array. """
        return np.dot(self.dyn_spec, self._dynconjtranspose) / (self.M - 1.0)  #ORDER OF OPERATIONS DEPENDENT (aka np.dot(t_dyn, dyn) doesn't work)


    @property
    def async_noscale(self):
        """ """
        return np.dot(self.dyn_spec, np.dot(self._noda, self._dynconjtranspose) ) / (self.M-1.0)


    @property
    def coeff_corr(self):
        """ Correlation coefficient (pg 78) """   
        return np.divide(self.sync_noscale, self.joint_var) 


    @property
    def coeff_disr(self):
        """ Disrelation coefficient (pg 79) """
        # Not the same as np.sqrt( 1 - coef_corr**2), only same in magnitude!
        return np.divide(self.async_noscale, self.joint_var)

    @property
    def joint_var(self):
        """ Product of standard devations of dynamic spectrum. 
        s1 * s2 or sqrt(siag(sync*sync)).
        """
        std = self.dyn_spec.std(axis=1) #sigma(lambda)
        return np.outer(std, std)
        #return Spec2d(np.outer(std, std),
                      #corr2d = self,
                      #name='Joint Variance',
                      #iunit='variance')


    # 11/10/14
    # I confirmed that these are equivalent to book definitions from 
    # diagonals of synchronous spectrum!  IE std = sqrt(diag(sync*sync))
    # and var = diag(sync * sync)
    # std is actually > var cuz var <1 so sqrt makes larger

    @property
    def _dynconjtranspose(self):
        """ Dynamic spectrum conjugate transpose; helpful to be cached"""
        return np.conj(self.dyn_spec).transpose()


    # 2D Correlation Spectra
    # ----------------------
    @property
    def sync(self):
        """ """
        if self._scaled:
            matrixout = self.sync_noscale * self.joint_var**(-1.0 * self.alpha) * \
                abs(self.coeff_corr)**(self.beta)
                    # ** faster than np.power but abs and np.abs same        
        else:
            matrixout = self.sync_noscale

        return Spec2d(matrixout, 
                      corr2d = self,
                      name='Synchronous Correlation',
                      iunit='synchronicity')   

    @property
    def async(self):
        """ """     
        if self._scaled:
            matrixout = self.async_noscale * self.joint_var**(-1.0 * self.alpha) * \
                abs(self.coeff_disr)**(self.beta)
        else:
            matrixout = self.async_noscale

        return Spec2d(matrixout, 
                      corr2d = self,
                      name='Asynchronous Correlation',
                      iunit='asynchronicity')   

    @property
    def phase(self):
        """ Global phase angle (pg 79).  This will use scaled data."""
        phase = np.arctan(self.async/self.sync)
        phase.name = 'Phase Map' 
        phase.iunit = 'phase angle'
        return phase    
    
    
    @property
    def modulous(self):
        """ Effective lengh the vector with components Sync/Async"""
        modulous = np.sqrt(self.sync**2 + self.async**2)
        modulous.name = 'Modulous'
        modulous.iunit = 'mod'
        return modulous
        

    @property
    def correlation(self):
        """ 2D Correlation Spectrum"""
        return Spec2d(self.coeff_corr, 
                      corr2d = self,
                      name = 'Correlation Coefficient',
                      iunit='corr. coefficient')                

    @property
    def disrelation(self):
        """ 2D Disrelation Spectrum"""
        return Spec2d(self.coeff_disr,
                      corr2d = self,
                      name = 'Disrelation Coefficient',
                      iunit='disr. coefficient')   


    # 2DCodistribution Spectroscopy
    # -----------------------------
    @property
    def char_index(self):
        """ Characteristic index.  In Ref. [2], this is the 
        characteristic time, and is equation 6.

        Returns: Spectrum of length equivalent to spectral index.
        """
        m = self.M

#        if self._center is None: #pre-center also has this case
        if np.count_nonzero(self.ref_spectrum) == 0:
            raise CorrError('CoDistribution divides by ref spectrum.  If'
                            ' not centring, the ref spec is 0 and you get infinities!')
        coeff = 1.0 / (m * self.ref_spectrum)

        summation = 0 
        k_matrix = np.empty(m)
        for k in range(1, m+1): #m+1 to include m in sum
            k_matrix.fill(k) # in place
            # df.mul is same as dotting:
                #http://stackoverflow.com/questions/15753916/dot-products-in-pandas
            summation += self.dyn_spec.dot(k_matrix) + ((m+1) / 2)
        return coeff * summation

    @property
    def char_perturb(self):
        """ Characteristic index.  In Ref. [2], this is the 
        characteristic time, and is equation 6.

        Returns: Spectrum of length equivalent to spectral index.
        """
        tm, t1 = self.columns[-1], self.columns[0]
        Kj = self.char_index
        
        return ((tm-t1) * ((Kj-1) / (self.M -1))) + t1


    # TO	VECTORIZE:
    
    @property
    def async_codist(self):
        """ Asynchronous codistribution """
        
        # Empty asyn matrix
        numrows = self.shape[0]        
        async = np.empty((numrows,numrows))

        tm, t1 = self.columns[-1], self.columns[0]
        
        # Numpy arrays to speed up loop/indexer?
        tbar = self.char_perturb.values
        var = self.joint_var
        
        # broadcast this?
        for i in range(numrows):
            for j in range(numrows):
                coeff = (tbar[j] - tbar[i]) / (tm -t1)
                # I believe std[i] std[j] is correct way
                async[i][j] = coeff * var[i,j]

        return Spec2d(async, 
                      corr2d=self, 
                      name='Asynchronous Codistribution', 
                      iunit='asynchronicity')
    
    @property
    def sync_codist(self):
        """ Syncrhonous codistribution.  Computed from asyn_codist"""
        numrows = self.shape[0]                    

        # Numpy arrays to speed up calculation
        var = self.joint_var
        async_cod = self.async_codist.values
        
        sync = np.empty((numrows,numrows))
        for i in range(numrows):
            for j in range(numrows):
                sync[i][j] = np.sqrt(var[i,j]**2 - async_cod[i,j]**2 )
                 
        return Spec2d(sync, 
                      corr2d=self, 
                      name='Synchronous Codistribution', 
                      iunit='synchronicity')
    

    def plot(self, **pltkwargs):
        """ Quad plot shows several kinds of correlation plots."""
        return corr_multi(self, **pltkwargs)
        

    def _pcagate(self, attr):
        """ Raise an error if use calls inaccessible PCA method."""
        if not self._PCA:
            raise CorrError('Please run .pca_fit() method before '
                            'calling %s.%s' % self.__class__.__name__, attr)    

    def pca_fit(self, n_components=None, fit_transform=True):# k=None, kernel=None, extern=False):           
        """         
        Adaptation of Alexis Mignon's pca.py script

        Adapted to fit skspec 5/6/2013.  
        Original credit to Alexis Mignon:
        Module for Principal Component Analysis.

        Author: Alexis Mignon (c)
        Date: 10/01/2012
        e-mail: alexis.mignon@gmail.com
        (https://code.google.com/p/pypca/source/browse/trunk/PCA.py)

        Constructor arguments:
        * k: number of principal components to compute. 'None'
             (default) means that all components are computed.
        * kernel: perform PCA on kernel matrices (default is False)
        * extern: use extern product to perform PCA (default is 
               False). Use this option when the number of samples
               is much smaller than the number of features.            

        See pca.py constructor for more info.

        This will initialize PCA class and fit current values of timespectra.

        Notes:
        ------
        The pcakernel.py module is more modular.  These class methods
        make it easier to perform PCA on a timespectra, but are less 
        flexible than using the module functions directly.

        timespectra gets transposed as PCA module expects rows as 
        samples and columns as features.

        Changes to timespectra do not retrigger PCA refresh.  This 
        method should be called each time changes are made to the data.
        """
        
        # NOW USES DYNSPEC BUT DID NOT TEST BEFORE CHANGING
        if self.center:
            logger.warn('Builtin PCA will perform mean-centering on'
                        ' data.  Data is not mean centered yet.')
        self._PCA = PCA(n_components=n_components)                
        if fit_transform:
            return self._PCA.fit_transform(self.dyn_spec)#.transpose())
        else:    
            self._PCA.fit(self.dyn_spec)#.transpose())


    @property
    def PCA(self):
        """ Return the full PCA class object"""
        self._pcagate('pca')
        return self._PCA 

    @property
    def pca_evals(self):
        self._pcagate('eigen values')
        # Index is not self.columns because eigenvalues are still computed with
        # all timepoints, not a subset of the columns        
        return self._PCA.eigen_values_

    @property
    def pca_evecs(self):
        self._pcagate('eigen vectors')
        return self._PCA.eigen_vectors_

    def load_vec(self, k):
        """ Return loading vector series for k.  If k > number of components
            computed with runpca(), this raises an error rather than 
            recomputing.
        """
        self._pcagate('load_vec')
        if k > len(self.shape[1]):
            raise CorrError('Principle components must be <= number'
                            'of samples %s'% self.shape[1])

        # Decided to put impetus on user to recompute when not using enough principle components
        # rather then trying to figure out logic of all use cases.
        # If k > currently stored eigenvectors, recomputes pca
        if self._PCA._k:
            if k > len(self.pca_evals):   
                logger.warn('Refitting, only %s components were computed'
                            'originally' % self._PCA._k)
                self.pca_fit(n_components=k, fit_transform=False)

        return self._PCA.eigen_vectors_[:,k]


    def __repr__(self):
        """ Aligned columns like pyparty.multicanvas """
        pad = pvconfig.PAD
        address = super(Corr2d, self).__repr__().split()[-1].strip("'").strip('>')

        outstring = '%s (%s X %s) at %s:\n' % (self.__class__.__name__,
                                               self.shape[0], self.shape[1], address)

        #Units
    #     outstring += '%sUnits -->  %s X %s\n' % (pad, self.specunit.lower(), self.varunit.lower())

        #Centering
        outstring += '%sCentering -->  %s\n' % (pad, self.center)

        #Scaling
        if self._scaled:
            outstring += '%sScaled    -->  %s\n' % (pad, self._scale_string)
        else:
            outstring += '%sScaled    -->  %s\n' % (pad, self._scaled)        

        outstring += '%sUnits     -->  [%s X %s]' % (pad, 
                                                     self.specunit.lower(), 
                                                     self.varunit.lower())
        return outstring


if __name__ == '__main__':
    from skspec.data import aunps_glass, solvent_evap, aunps_water
    import numpy as np
    import matplotlib.pyplot as plt 

#    ts=aunps_water().as_varunit('s')
    ts = solvent_evap()#.as_varunit('s').as_norm('r')

    cd = Corr2d(ts)#, refspec=ts.mean(axis=1) )
    cd.set_center('mean')
    print cd.sync
    cd.scale(alpha=0, beta=0)
    cd.sync.plot(contours=20, kind='corr3d')
    plt.show()
#    cd.scale(a=.5, b=.5)
    
#    cd.sync
#    cd.async_noscale



import numpy as np
from chaco.api import AbstractPlotData
from traits.api import Any, Dict, Tuple
import collections

import time #testing

class PandasPlotData(AbstractPlotData):

    ''' Chaco requires a PlotData interface to manage plot/data mapping; however, pandas
    already is its own nice data handler, so this class mostly provides a wrapper that tries
    to be a hidden middle man.  As such, I try to make this PandasPlotData feel like a pandas
    df.  It would be nice to have chaco communicate with a df directly, but for now
    it must do so through this liason class.'''

    df=Any()
    
    # Used to store actual column values and string representation
    _colmap=Dict()
    _extras=Dict()
    _reserved=Tuple('index', 'columns')
    
    # Dict mapping data series name to a mask array holding the selections
    selections = Dict()    
    
     

    def __init__(self, df, extras=None, **kwtraits):
        """PandasPlotData exposes a PlotData interface from a df.
        It is chosen that this object MUST be initialized with a df.
        """

        super(PandasPlotData, self).__init__(**kwtraits) #Trait initialization 
        
        if len(df.shape) > 2:
            raise NotImplementedError('Multidimensional dfs of order 3 or higher \
	    are not supported by in PandasPlotData') #Do 1d arrays work?
        
        self.set_df(df)

    #------------------------------------------------------------------------
    # Dictionary Interface to PandasPlot Data
    #------------------------------------------------------------------------

    def __getitem__(self, name):
        return self.df[name]

    def __setitem__(self, name, value):
        raise NotImplemented
        
    def __delitem__(self, name):
        raise NotImplemented           

    def list_data(self, as_strings=False):
        """ Returns a list of the names of the selections managed by this instance. 
 
            as_strings: If true, names are converted to strings.  Useful for passing 
                        directly into getdata in which strings are stored.
        """
        if self.df is None:
            return []        
        if as_strings:
            return [str(col) for col in self.df.columns]
        else:
            return list(self.df.columns.values)

    def get_data(self, name):
        ''' Must be called via string name because Plot() enforces that all stored
            datasources are accessed via string!.'''
        
        

        
        ### If not a string, immediately return dataframe value
        if not isinstance(name, basestring):
            return self.df[name].values


        else:

            ### Should do test for reserved names to avoid conflicts here
            if name=='index':
                return self.df.index.values
            
            elif name=='columns':
                return self.df.columns.values            
            
            ### Try to get value from stringmap to dataframe columns
            try:
                return self.df[self._colmap[name]].values
            except KeyError:
                pass
            
            ### Try to get value from extras 
            try:
                return self._extras[name]
            except KeyError:
                raise AttributeError('%s not found in pandasplotdata.  Please check list_data() for valid entries.'%name)
        
                
            
    def del_data(self, name):
        """ Deletes the array specified by *name*, or raises a KeyError if
        the named array does not exist.
        """
        raise NotImplementedError('Do I want to delete data from a df?')
                                


    ### SHOULD TIME TEST TO SEE HOW LONG IS TAKES
    def set_data(self, name, new_data):
        """ Takes in name/array data similar to arrayplotdata.  Unlike arrayplotdata,
        this never adds values to the self.arrays.  Unlike arrayplotdata, this is not
        willing to accept new entries.  It is up to the Plot object pass in the correct
        data.
        """
        if not self.writable:
            return None

        try:
            new_data = np.array(new_data) #Convert to array data
        except Exception:
            raise Exception('set_data failed to convert type %s data to array'%type(new_data))

                
        if not isinstance(name, basestring):
            raise TypeError('set_data() can only accept string keys.')
                
        
        if name in self._stringmap:
            raise AttributeError('%s cannot correspond to a string name already present in \
                                  in the underlying dataframe.'%name)
            
        if name in self._reserved:
            raise NameError('%s conflicts with pandasplotdata reserved names: "%s"'%(name, ','.join(self._reserved)))
 
                   
        if name in self._extras:
            self.data_changed = {'changed':[name]}       
        else:
            self.data_changed = {'added':[name]}       

        self._extras[name]=new_data

        return name #THIS MUST BE HERE (why?)

    
    def set_df(self, dfnew):
        ''' Set an entirely new dataframe.  Calls _set_df() to handle event changes.  
            Returns T/F to let plot know if event was "changed" vs. added or removed.
            _set_df() prevents mixed events of changed + added/removed.'''
        
        event=self._set_df(dfnew)
        event2=self._reset_extras()

        if event2:
            event.update(event2)
            
        self.data_changed=event
        
        if event.has_key('changed'): #Changed will be the only call.
            return True
        return False
    
           
    def _set_df(self, dfnew):
        ''' Handles add/remove/change events for a new dataframe.'''
        
        event={}
        
        ### If self.df not initialized, add all entries
        if isinstance(self.df, type(None)):
            pass
    
        else:
            ### If only column names are changing, trigger change event and exit
            if np.array_equal(dfnew.columns.values, self.df.columns.values): #very fast operation
                self.df=dfnew
                event['changed']=self._colmap.keys()
                return event
                
            else:
                ### Remove old columns
                event['removed']=self._colmap.keys()

        ### Add new columns/update self.df
        event.update(self._add_df(dfnew))  
            
        return event 
        
        ### MORE COMPLEX VERSION ANY FASTER? ###
#        if not np.array_equal(dfnew.columns.values, self.df.columns.values):
    ##        raise IOError('set_df may only be called for objects of identical columns.')
            #event={'removed':self._colmap.keys()}
            #event['added']=dfnew.columns.values
    
        
        #else:
            #self.df=dfnew
            #event = {'changed':dfnew.columns.values}       
           
        #self.data_changed=event         
        
    def _reset_extras(self):
        if self._extras:
            return {'removed':self._extras.keys()}
        
    def _add_df(self, dfnew):
        ''' Add all columns of new df, update colmap, set self.df'''
        self._colmap=dict((str(c), c) for c in dfnew.columns.values)        
        self.df=dfnew        
        return {'added':self._colmap.keys()}
                   
    def _get_indicies(self, *names):
        ''' Takes in a list of names and returns indicies corresponding to.  Useful for label
        mappers.'''
        raise NotImplemented
#        return [self.idxget(str(name)) for name in names]

    ### These are used by chaco to inform the plot that a certain region of data is selected
    def get_selection(self, name):
        """ Returns the selection for the given column name """
        print 'hi being selected in plotdata'
        return self.selections.get(name, None)

    def set_selection(self, name, selection):
        # Store the selection in a separate dict mapping name to its
        # selection array
        self.selections[name] = selection
        self.data_changed = True
